@@ -23,6 +23,7 @@ ffi_type * MVM_nativecall_get_ffi_type(MVMThreadContext *tc, MVMuint64 type_id)
         case MVM_NATIVECALL_ARG_ASCIISTR:
         case MVM_NATIVECALL_ARG_UTF8STR:
         case MVM_NATIVECALL_ARG_UTF16STR:
+        case MVM_NATIVECALL_ARG_CPPSTRUCT:
         case MVM_NATIVECALL_ARG_CSTRUCT:
         case MVM_NATIVECALL_ARG_CPOINTER:
         case MVM_NATIVECALL_ARG_CARRAY:
@@ -81,7 +82,6 @@ static void * unmarshal_callback(MVMThreadContext *tc, MVMObject *callback, MVMO
     /* Try to locate existing cached callback info. */
     callback = MVM_frame_find_invokee(tc, callback, NULL);
     cuid     = ((MVMCode *)callback)->body.sf->body.cuuid;
-    MVM_string_flatten(tc, cuid);
     MVM_HASH_GET(tc, tc->native_callback_cache, cuid, callback_data_head);
 
     if (!callback_data_head) {
@@ -207,12 +207,18 @@ static void callback_handler(ffi_cif *cif, void *cb_result, void **cb_args, void
     CallbackInvokeData cid;
     MVMint32 num_roots, i;
     MVMRegister res;
+    MVMRegister *args;
     MVMNativeCallback *data = (MVMNativeCallback *)cb_data;
     void           **values = MVM_malloc(sizeof(void *) * (data->cs->arg_count ? data->cs->arg_count : 1));
 
-    /* Build a callsite and arguments buffer. */
+    /* Unblock GC if needed, so this thread can do work. */
     MVMThreadContext *tc = data->tc;
-    MVMRegister    *args = MVM_malloc(data->num_types * sizeof(MVMRegister));
+    MVMint32 was_blocked = MVM_gc_is_thread_blocked(tc);
+    if (was_blocked)
+        MVM_gc_mark_thread_unblocked(tc);
+
+    /* Build a callsite and arguments buffer. */
+    args = MVM_malloc(data->num_types * sizeof(MVMRegister));
     num_roots = 0;
     for (i = 1; i < data->num_types; i++) {
         MVMObject *type     = data->types[i];
@@ -407,6 +413,10 @@ static void callback_handler(ffi_cif *cif, void *cb_result, void **cb_args, void
     MVM_gc_root_temp_pop_n(tc, num_roots);
     MVM_free(args);
     MVM_free(cif);
+
+    /* Re-block GC if needed, so other threads will be able to collect. */
+    if (was_blocked)
+        MVM_gc_mark_thread_blocked(tc);
 }
 
 #define handle_arg(what, cont_X, dc_type, reg_slot, unmarshal_fun) do { \
@@ -435,6 +445,21 @@ static void callback_handler(ffi_cif *cif, void *cb_result, void **cb_args, void
     } \
 } while (0)
 
+#define handle_ret(tc, c_type, ffi_type, make_fun) do { \
+    if (sizeof(c_type) < sizeof(ffi_type)) { \
+        ffi_type ret; \
+        ffi_call(&cif, entry_point, &ret, values); \
+        MVM_gc_mark_thread_unblocked(tc); \
+        result = make_fun(tc, res_type, (c_type)ret); \
+    } \
+    else { \
+        c_type ret; \
+        ffi_call(&cif, entry_point, &ret, values); \
+        MVM_gc_mark_thread_unblocked(tc); \
+        result = make_fun(tc, res_type, ret); \
+    } \
+} while (0)
+
 MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
         MVMObject *site, MVMObject *args) {
     MVMObject     *result = NULL;
@@ -498,6 +523,22 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
                 values[i]           = MVM_malloc(sizeof(void *));
                 *(void **)values[i] = MVM_nativecall_unmarshal_cstruct(tc, value);
                 break;
+            case MVM_NATIVECALL_ARG_CPPSTRUCT: {
+                /* We need to allocate the struct (THIS) for C++ constructor before passing it along. */
+                if (i == 0 && !IS_CONCRETE(value)) {
+                    MVMCPPStructREPRData *repr_data = (MVMCPPStructREPRData *)STABLE(res_type)->REPR_data;
+                    /* Allocate a full byte aligned area where the C++ structure fits into. */
+                    void *ptr           = MVM_malloc(repr_data->struct_size > 0 ? repr_data->struct_size : 1);
+                    result              = MVM_nativecall_make_cppstruct(tc, res_type, ptr);
+                    values[i]           = MVM_malloc(sizeof(void *));
+                    *(void **)values[i] = ptr;
+                }
+                else {
+                    values[i]           = MVM_malloc(sizeof(void *));
+                    *(void **)values[i] = MVM_nativecall_unmarshal_cppstruct(tc, value);
+                }
+                break;
+            }
             case MVM_NATIVECALL_ARG_CPOINTER:
                 if ((arg_types[i] & MVM_NATIVECALL_ARG_RW_MASK) == MVM_NATIVECALL_ARG_RW) {
                     values[i]                     = MVM_malloc(sizeof(void *));
@@ -546,130 +587,107 @@ MVMObject * MVM_nativecall_invoke(MVMThreadContext *tc, MVMObject *res_type,
     }
 
 
-    /* Process return values. */
     MVMROOT(tc, args, {
     MVMROOT(tc, res_type, {
-        switch (ret_type & MVM_NATIVECALL_ARG_TYPE_MASK) {
-            case MVM_NATIVECALL_ARG_VOID:
-                ffi_call(&cif, entry_point, values[body->num_args], values);
-                result = res_type;
-                break;
-            case MVM_NATIVECALL_ARG_CHAR: {
-                signed char ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
+        MVM_gc_mark_thread_blocked(tc);
+        if (result) {
+            /* We are calling a C++ constructor so we hand back the invocant (THIS) we recorded earlier. */
+            void *ret; // We are not going to use it, but we need to pass it to libffi.
+            ffi_call(&cif, entry_point, &ret, values);
+            MVM_gc_mark_thread_unblocked(tc);
+        }
+        else {
+            /* Process return values. */
+            switch (ret_type & MVM_NATIVECALL_ARG_TYPE_MASK) {
+                case MVM_NATIVECALL_ARG_VOID: {
+                    void *ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    MVM_gc_mark_thread_unblocked(tc);
+                    result = res_type;
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_CHAR:
+                    handle_ret(tc, signed char, ffi_sarg, MVM_nativecall_make_int);
+                    break;
+                case MVM_NATIVECALL_ARG_SHORT:
+                    handle_ret(tc, signed short, ffi_sarg, MVM_nativecall_make_int);
+                    break;
+                case MVM_NATIVECALL_ARG_INT:
+                    handle_ret(tc, signed int, ffi_sarg, MVM_nativecall_make_int);
+                    break;
+                case MVM_NATIVECALL_ARG_LONG:
+                    handle_ret(tc, signed long, ffi_sarg, MVM_nativecall_make_int);
+                    break;
+                case MVM_NATIVECALL_ARG_LONGLONG:
+                    handle_ret(tc, signed long long, ffi_sarg, MVM_nativecall_make_int);
+                    break;
+                case MVM_NATIVECALL_ARG_FLOAT: {
+                    float ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    MVM_gc_mark_thread_unblocked(tc);
+                    result = MVM_nativecall_make_num(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_DOUBLE: {
+                    double ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    MVM_gc_mark_thread_unblocked(tc);
+                    result = MVM_nativecall_make_num(tc, res_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_ASCIISTR:
+                case MVM_NATIVECALL_ARG_UTF8STR:
+                case MVM_NATIVECALL_ARG_UTF16STR: {
+                    char *ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    MVM_gc_mark_thread_unblocked(tc);
+                    result = MVM_nativecall_make_str(tc, res_type, body->ret_type, ret);
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_CSTRUCT:
+                    handle_ret(tc, void *, ffi_arg, MVM_nativecall_make_cstruct);
+                    break;
+                case MVM_NATIVECALL_ARG_CPPSTRUCT:
+                    handle_ret(tc, void *, ffi_arg, MVM_nativecall_make_cppstruct);
+                    break;
+                case MVM_NATIVECALL_ARG_CPOINTER:
+                    handle_ret(tc, void *, ffi_arg, MVM_nativecall_make_cpointer);
+                    break;
+                case MVM_NATIVECALL_ARG_CARRAY:
+                    handle_ret(tc, void *, ffi_arg, MVM_nativecall_make_carray);
+                    break;
+                case MVM_NATIVECALL_ARG_CUNION:
+                    handle_ret(tc, void *, ffi_arg, MVM_nativecall_make_cunion);
+                    break;
+                case MVM_NATIVECALL_ARG_CALLBACK: {
+                    /* TODO: A callback -return- value means that we have a C method
+                    * that needs to be wrapped similarly to a is native(...) Perl 6
+                    * sub. */
+                    void *ret;
+                    ffi_call(&cif, entry_point, &ret, values);
+                    MVM_gc_mark_thread_unblocked(tc);
+                    /* XXX do something with the function pointer: ret */
+                    result = res_type;
+                    break;
+                }
+                case MVM_NATIVECALL_ARG_UCHAR:
+                    handle_ret(tc, unsigned char, ffi_arg, MVM_nativecall_make_int);
+                    break;
+                case MVM_NATIVECALL_ARG_USHORT:
+                    handle_ret(tc, unsigned short, ffi_arg, MVM_nativecall_make_int);
+                    break;
+                case MVM_NATIVECALL_ARG_UINT:
+                    handle_ret(tc, unsigned int, ffi_arg, MVM_nativecall_make_int);
+                    break;
+                case MVM_NATIVECALL_ARG_ULONG:
+                    handle_ret(tc, unsigned long, ffi_arg, MVM_nativecall_make_int);
+                    break;
+                case MVM_NATIVECALL_ARG_ULONGLONG:
+                    handle_ret(tc, unsigned long long, ffi_arg, MVM_nativecall_make_int);
+                    break;
+                default:
+                    MVM_exception_throw_adhoc(tc, "Internal error: unhandled libffi return type");
             }
-            case MVM_NATIVECALL_ARG_SHORT: {
-                signed short ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_INT: {
-                signed int ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_LONG: {
-                signed long ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_LONGLONG: {
-                signed long long ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_FLOAT: {
-                float ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_num(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_DOUBLE: {
-                double ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_num(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_ASCIISTR:
-            case MVM_NATIVECALL_ARG_UTF8STR:
-            case MVM_NATIVECALL_ARG_UTF16STR: {
-                char *ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_str(tc, res_type, body->ret_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_CSTRUCT: {
-                void *ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_cstruct(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_CPOINTER: {
-                void *ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_cpointer(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_CARRAY: {
-                void *ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_carray(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_CUNION: {
-                void *ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_cunion(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_CALLBACK: {
-                /* TODO: A callback -return- value means that we have a C method
-                * that needs to be wrapped similarly to a is native(...) Perl 6
-                * sub. */
-                void *ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                /* XXX do something with the function pointer: ret */
-                result = res_type;
-                break;
-            }
-            case MVM_NATIVECALL_ARG_UCHAR: {
-                unsigned char ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_USHORT: {
-                unsigned short ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_UINT: {
-                unsigned int ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_ULONG: {
-                unsigned long ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            case MVM_NATIVECALL_ARG_ULONGLONG: {
-                unsigned long long ret;
-                ffi_call(&cif, entry_point, &ret, values);
-                result = MVM_nativecall_make_int(tc, res_type, ret);
-                break;
-            }
-            default:
-                MVM_exception_throw_adhoc(tc, "Internal error: unhandled libffi return type");
         }
     });
     });
